 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */

package Java2OWL;

import java.util.*;
import org.semanticweb.owlapi.model.OWLClass;

/** This class maintains the correlation between a Java class and an OWL class. 
 * It stores the Java class and the corresponding OWL class together with the {@see PropertyWrapper}s.
 * They maintain the correlation between Java-methods and OWL-properties.
 *
 * @author ohlbach
 */
public class ClassWrapper {
    
    /** the top manager */
    J2OManager manager; 
    
    /** the Java class. */
    private Class jClass;
   
    /** the OWL class. */
    private OWLClass oClass;
 
    /** the PropertyWrapper for all the getter-methods in the Java class which are mapped to OWL properties. */
    private ArrayList<PropertyWrapper> properties = new ArrayList<PropertyWrapper>();
   
    /** constructs a new ClassWrapper instance.
    * 
    * @param jClass a Java-class.
    * @param oClass the corresponding OWL-class.
    */ 
    ClassWrapper(Class jClass, OWLClass oClass,J2OManager manager) {
        this.jClass = jClass;
        this.oClass = oClass;
        this.manager = manager;}
    
    /** adds a PropertyWrapper. */
    void addProperty(PropertyWrapper property) {properties.add(property);}

    /** returns the Java class. */
    public Class getJavaClass() {return jClass;} 

    /** returns the OWL-class. */
    public OWLClass getOWLClass() {return oClass;}
    
    /** returns the list of PropertyWrappers. */
    ArrayList<PropertyWrapper> getProperties() {return properties;}
    
    /** returns a short name for the OWL-Class. */ 
    String oClassShortName() {return oClass.getIRI().getFragment();}
    
    /** returns a short name for the Java-Class. */ 
    String jClassShortName() {return jClass.getSimpleName();}
    
    /** returns the mapping as a string JavaClass <-> OWLClass. */
    public String getName() {
        return jClass.getSimpleName() + " <-> " + oClass.getIRI().getFragment();}
    
    /** instantiates the classes in the ClassWrappers and puts them into the objects-array.
     * 
     * @param clwrps a collection of ClassWrappers.
     * @param objects an Object array of sufficient length.
     * @param i the index from where on the new objects are to be put into the array.
     * @param logger a logger.
     * @return true if the operation succeeded.
     * <p> An instantiation error causes an error message to be logged and the method to stop immediately.
     */
    static boolean instantiate(Collection<ClassWrapper> clwrps, Object[] objects, int i, Logging logger) {
        for(ClassWrapper clw : clwrps) {
            Object object = null;
            try{object = clw.jClass.newInstance();}
            catch(Exception ex){
                logger.error("Class instantiation error", Logging.Origin.USAGE, clw.jClass,ex);
                return false;}
            objects[i] = object;
            logger.event(Logging.Semantics.JOC,object.getClass(),object.hashCode());
            ++i;}
        return true;}
    
    /** returns a description of the ClassWrapper */
    @Override
    public String toString() {
        String s = jClass.getSimpleName() + " <-> ";
        s += (oClass == null) ? "null" : oClass.getIRI().getFragment();
        if(!properties.isEmpty()) {
            s += "\nProperty Wrapper: ";
            for(PropertyWrapper pmd : properties) s += pmd.getName() + ", ";
            s = s.substring(0,s.length()-2);}
        return s;}}

