 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */
package Java2OWL;

import java.io.File;
import java.util.*;
import org.semanticweb.HermiT.Reasoner;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.*;
import com.clarkparsia.pellet.owlapiv3.*;
import java.lang.reflect.Method;
import java.util.Map.Entry;
import java.util.logging.Level;
import org.coode.owlapi.manchesterowlsyntax.ManchesterOWLSyntaxEditorParser;
import org.semanticweb.owlapi.expression.*;
import org.semanticweb.owlapi.util.*;
import org.semanticweb.owlapi.vocab.OWL2Datatype;
import uk.ac.manchester.cs.factplusplus.owlapiv3.*;

/** This class manages the ontologies for the Java2OWL-System.
 * <p>
 * At compile time, there are usually two ontologies, a background ontology (which may be empty)
 * and an extension ontology. The extension ontology may be filled with translated Java classes and methods.
 * At run time there is the extension ontology or an extension of the original extension ontology.
 * <p>
 * Although there may be different J2OOntologyManagers with different ontologies in parallel,
 * there is only one global OWL-ontology manager and OWL data factory.
 * <p>
 * The ontology manager can be in debug mode or run mode (debugMode = false).
 * In debug mode all changes to the ontology are immediately forwarded to the reasoner
 * and consistency is checked immediately. 
 * In run mode changes to the ontology are cached until the ontology is accessed in some way. 
 * <p>
 * Notice that most of private variables are not set by the constructor, but when an ontology is loaded.
 * 
 * <p>
 * <strong>Thread Safety</strong><br/>
 * All methods which are needed only during compilation and linking phase are not thread safe.</p>
 * The methods needed during the application phase are synchronised with changeAxioms as monitor. 
 * The changeAxioms are the interface to the ontology. 
 * 
 * @author ohlbach
 */
public class J2OOntologyManager {
    
    /** the top manager. */ 
    private J2OManager manager; 
    
    /** the class manager */
    private J2OClassManager classManager;
    
    /** sets the class manager */
    void setClassManager(J2OClassManager clm) {classManager = clm;}
    
    /** indicates the debug mode. */
    private boolean debugMode = false;
    
    /** returns the name of the ontology manager.*/
    public String getName() {return manager.getName() + "_OntologyManager";}
    
    /** an OWLontology manager. */
    final OWLOntologyManager ontologyManager = OWLManager.createOWLOntologyManager();
   
    /** returns the OWL ontology manager (for advances purposes). */
    public OWLOntologyManager getOWLOntologyManager() {return ontologyManager;}
    
    /** a global data factory. */
    static final OWLDataFactory dataFactory = OWLManager.getOWLDataFactory();
    
    /** maps Java classes to OWLDatatypes. */
    static final Map<Class,OWLDatatype> toOWLDatatype = new HashMap<Class,OWLDatatype>();
    
    static {
            OWLDatatype dt = dataFactory.getOWLDatatype(IRI.create("http://www.w3.org/2001/XMLSchema#byte"));
            toOWLDatatype.put(Byte.class,dt);       toOWLDatatype.put(byte.class,dt);
            dt = dataFactory.getOWLDatatype(IRI.create("http://www.w3.org/2001/XMLSchema#short"));
            toOWLDatatype.put(Short.class,dt);      toOWLDatatype.put(short.class,dt);
            dt = dataFactory.getOWLDatatype(IRI.create("http://www.w3.org/2001/XMLSchema#unsignedShort"));
            toOWLDatatype.put(Character.class,dt);  toOWLDatatype.put(char.class,dt);
            dt = dataFactory.getIntegerOWLDatatype();
            toOWLDatatype.put(Integer.class,dt);    toOWLDatatype.put(int.class,dt);
            dt = dataFactory.getOWLDatatype(IRI.create("http://www.w3.org/2001/XMLSchema#long"));
            toOWLDatatype.put(Long.class,dt);       toOWLDatatype.put(long.class,dt);
            dt = dataFactory.getFloatOWLDatatype();
            toOWLDatatype.put(Float.class,dt);      toOWLDatatype.put(float.class,dt);
            dt = dataFactory.getDoubleOWLDatatype();
            toOWLDatatype.put(Double.class,dt);     toOWLDatatype.put(double.class,dt);
            dt = dataFactory.getBooleanOWLDatatype();
            toOWLDatatype.put(Boolean.class,dt);    toOWLDatatype.put(boolean.class,dt); 
            dt = dataFactory.getOWLDatatype(IRI.create("http://www.w3.org/2001/XMLSchema#string"));
            toOWLDatatype.put(String.class,dt);
    }
    
    /** returns the datatype mapping as a string, just for testing purposes.*/
    private static String datatypeMappingString(String prefix) {
        String s = "";
        for(Entry<Class,OWLDatatype> entry : toOWLDatatype.entrySet()) {
            s += prefix +entry.getKey().getSimpleName() + " -> " + entry.getValue().toString() + "\n";}
        return Utilities.adjustPairs(s, "->");}
    
    /** to create an annotation for generated OWL-classes that refer to the corresponding class name. */
    private OWLAnnotationProperty classAnnotationProperty = null;
    
    /** to create an annotation for generated OWL-properties that refer to the corresponding method name. */
    private OWLAnnotationProperty methodAnnotationProperty = null;
    
    /** to create an annotation for generated OWL-properties that refer to the corresponding class. */
    private OWLAnnotationProperty classesAnnotationProperty = null;
    
    /** returns the change axioms as string of lines (for test purposes). */
    String changeAxiomsString() {
        String s = "";
        for(OWLOntologyChange changeAxiom : changeAxioms) 
            s += changeAxioms.toString()+"\n";
        return s;}
  
    /** logs events. */
    private Logging logger;
    
    /** creates a new J2OOntologyManager for the J2OManager. 
     The remaining private variables are set when an ontology is loaded.*/ 
    J2OOntologyManager(J2OManager manager) {
        this.manager = manager;
        debugMode = manager.isInDebugMode();
        logger = manager.getLogger();
        logger.installation(this);}
    
    
    //--------------------- Begin: OWL Ontology Management ---------------------//
    
    /** the background ontology at compile and run time. */
    private OWLOntology         backgroundOntology = null;
    
    /** the extension ontology in all system states. */
    private OWLOntology         extensionOntology = null;
       
    /** the IRI of the extension ontology. */
    private String extensionOntologyIRI = "";
    
    /** an entity remover for the extension ontology. */
    private OWLEntityRemover remover = null;
    
    /** maps OWL-Entities to names. */
    private BidirectionalShortFormProvider bidiShortFormProvider = null;
    
    /** maps strings to OWL entities.*/
    private OWLEntityChecker entityChecker = null; 

    /** returns the background ontology. */
    public OWLOntology getBackgroundOntology() {return backgroundOntology;}
 
    /** returns the extension ontology. */
    public OWLOntology getExtensionOntology() {return extensionOntology;}
    
    /** returns the IRI of the extension ontology (or null if there is in no extension ontology). 
     */
    IRI getExtensionOntologyIRI() {
        if(extensionOntology != null) return extensionOntology.getOntologyID().getDefaultDocumentIRI();
        else return null;}
    
    /** returns the IRI of the background ontology (or null if there is in no extension ontology). */
    IRI getBackgroundOntologyIRI() {
        if(backgroundOntology != null) return backgroundOntology.getOntologyID().getDefaultDocumentIRI();
        else return null;}
    
    /** adds an IRIMapper to the ontology manager. 
     The IRIMapper helps finding imported ontologies. */
    public void addIRIMapper(AutoIRIMapper iriMapper) {
          ontologyManager.addIRIMapper(iriMapper);
          logger.installation(iriMapper);}
    
    /** loads an ontology from an IRI. 
     * 
     * <p>
     * A previously installed ontology is removed (and not saved).
     * 
     * @param iri the document IRI of the ontology.
     * @param background if true, then the background ontology is loaded, otherwise the extension ontology.
     * @return the loaded ontology or null if something goes wrong.
     */
    public OWLOntology loadOntology(IRI iri, boolean background) {
        unloadOntology(background);
        OWLOntology ontology = null;
        try {ontology = ontologyManager.loadOntologyFromOntologyDocument(iri);} 
        catch (OWLOntologyCreationException ex) {
            logger.error("OntologyCreationException",Logging.Origin.USAGE,iri,ex);
            return null;}
        if(background) this.backgroundOntology = ontology;
        else {
            this.extensionOntology = ontology;
            remover = new OWLEntityRemover(ontologyManager,extensionOntology.getImportsClosure());
            extensionOntologyIRI = ontology.getOntologyID().getDefaultDocumentIRI().toString();
            classAnnotationProperty = dataFactory.getOWLAnnotationProperty(IRI.create(extensionOntologyIRI+"#Java-Class"));
            methodAnnotationProperty = dataFactory.getOWLAnnotationProperty(IRI.create(extensionOntologyIRI+"#Java-Method"));
            classesAnnotationProperty = dataFactory.getOWLAnnotationProperty(IRI.create(extensionOntologyIRI+"#Java-Classes"));
            loadReasoner();}

        // The rest is needed for the Manchaster Syntax parser.
        if(bidiShortFormProvider != null) bidiShortFormProvider.dispose();
        Set<OWLOntology> importsClosure = ontology.getImportsClosure();
        ShortFormProvider shortFormProvider = new SimpleShortFormProvider();
        bidiShortFormProvider = new BidirectionalShortFormProviderAdapter(ontologyManager, importsClosure, shortFormProvider);
        entityChecker =  new ShortFormEntityChecker(bidiShortFormProvider);
        logger.info("Ontology installed", ontology);
        return ontology;}
   
    /** loads the background ontology.
     * 
     * @param name the place to load the ontology from:
     * - either an IRI as string, stating with http: or file:
     * - or a pathname of a file where the ontology is stored. 
     * @return the loaded ontology or null if it could not be loaded.<br/>
    */
    public OWLOntology loadBackgroundOntology(String name) {
        if(name == null || name.isEmpty()) return null;
        IRI iri = (name.startsWith("http:") || name.startsWith("file:")) ? IRI.create(name) : IRI.create(new File(name));
        return loadOntology(iri,true);}
    
    
    /** loads the extension ontology.
     * 
     * @param name the place to load the ontology from:
     * - either an IRI as string, stating with http: or file:
     * - or a pathname of a file where the ontology is stored. 
     * @return the loaded ontology or null if it could not be loaded.<br/>
    */
    public OWLOntology loadExtensionOntology(String name) {
        if(name == null || name.isEmpty()) return null;
        IRI iri = (name.startsWith("http:") || name.startsWith("file:")) ? IRI.create(name) : IRI.create(new File(name));
        return loadOntology(iri,false);}
    
    
    /** unloads the background ontology or the extension ontology. */
    void unloadOntology(boolean background) {
        if(background && this.backgroundOntology != null) {
            ontologyManager.removeOntology(this.backgroundOntology);
            logger.info("Ontology removed", this.backgroundOntology);
            this.backgroundOntology = null;}
        if(!background && this.extensionOntology != null) {
            ontologyManager.removeOntology(this.extensionOntology);
            logger.info("Ontology removed", this.extensionOntology);
            this.extensionOntology = null;} }
        
    /** This method opens an extension of the background ontology to be filled with new concepts 
     * translated from Java classes.
     * <p>
     * If importBO = true then the extension ontology refers with an import statement to the background ontology.<br/>
     * If importBO = false then the background ontology is copied to the extension ontology.
     *
     * @param iri is the IRI of the new extension ontology.
     * @param importBO determines whether the ontology extension should import the background ontology or use a copy.
     * @return the extension ontology, or null if something goes wrong.
     */
    OWLOntology createExtensionOntology(boolean importBO, IRI iri)  {
        if(extensionOntology != null) unloadOntology(false);
        changeAxioms.clear();
        extensionOntology = null;
        try {
            if(importBO) {
                this.extensionOntology = ontologyManager.createOntology(iri);
                if(backgroundOntology != null) {
                    changeAxioms.add(new AddImport(extensionOntology, dataFactory.getOWLImportsDeclaration(getBackgroundOntologyIRI())));
                    ontologyManager.applyChanges(changeAxioms);
                    changeAxioms.clear();
                }}
            else {
                Set<OWLOntology> imported = new HashSet<OWLOntology>();
                if(backgroundOntology != null) imported.add(backgroundOntology);
                this.extensionOntology = ontologyManager.createOntology(iri,imported);}}
        catch(OWLOntologyCreationException ex){
            logger.error("OntologyCreationException",Logging.Origin.USAGE,iri);
            return null; }
            
        Set<OWLOntology> importsClosure = extensionOntology.getImportsClosure();
        ShortFormProvider shortFormProvider = new SimpleShortFormProvider();
        if(bidiShortFormProvider != null) bidiShortFormProvider.dispose();
        bidiShortFormProvider = new BidirectionalShortFormProviderAdapter(ontologyManager, importsClosure, shortFormProvider);
        entityChecker =  new ShortFormEntityChecker(bidiShortFormProvider);
        extensionOntologyIRI = extensionOntology.getOntologyID().getDefaultDocumentIRI().toString();
        classAnnotationProperty = dataFactory.getOWLAnnotationProperty(IRI.create(extensionOntologyIRI+"#Java-Class"));
        methodAnnotationProperty = dataFactory.getOWLAnnotationProperty(IRI.create(extensionOntologyIRI+"#Java-Method"));
        classesAnnotationProperty = dataFactory.getOWLAnnotationProperty(IRI.create(extensionOntologyIRI+"#Java-Classes"));
        remover = new OWLEntityRemover(ontologyManager,extensionOntology.getImportsClosure());
        loadReasoner();
        logger.info("Ontology installed", extensionOntology);
        return extensionOntology;}
    
    /** saves the extension ontology either to the previously specified IRI, or to the given file. 
     * 
     * @param iri the IRI where the extension ontology is to be saved.
     * @return true if the operation succeeded. */
    public boolean saveExtensionOntology(IRI iri) {
       if(extensionOntology == null) 
            {logger.error("Extension Ontology undefined.",Logging.Origin.USAGE); return false;}
        applyChanges(true);
        try {
            ontologyManager.saveOntology(extensionOntology, iri);
            logger.info("Ontology saved",extensionOntology.toString() + " to file " + iri.toString());}
        catch(OWLOntologyStorageException ex) {
            logger.error("OntologyStorageException",Logging.Origin.USAGE,ex);
            return false;}
        return true;}

    
    /** saves the extension ontology either to the previously specified IRI, or to the given file. 
     * @param filename optionally a filename where the ontology is to be stored.
     * @return true if the operation succeeded. */
    public boolean saveExtensionOntology(String filename) {
        if(extensionOntology == null)
            {logger.error("Extension Ontology undefined.",Logging.Origin.USAGE); return false;}
        applyChanges(true);
        try {
            if(filename != null && !filename.isEmpty()) {
                ontologyManager.saveOntology(extensionOntology, IRI.create(new File(filename)));
                logger.info("Ontology saved",extensionOntology.toString() + "into file " +filename);}
            else {
                ontologyManager.saveOntology(extensionOntology);
                logger.info(Level.INFO,"Ontology saved",extensionOntology);}}
        catch(OWLOntologyStorageException ex){
            logger.error("OntologyStorageException",Logging.Origin.USAGE,ex);
            return false;}
        return true;}
    
    /** saves the extension ontology to the same file where it was loaded from 
    * <p>OWLOntologyStorageExceptions are logged.
    * 
    * @return true if the operation succeeded. */
    public boolean saveExtensionOntology() {
        if(extensionOntology == null) 
            {logger.error("Extension Ontology undefined.",Logging.Origin.USAGE); return false;}
        applyChanges(true);
        try {
             ontologyManager.saveOntology(extensionOntology);
             logger.info("Ontology saved",extensionOntology);}
        catch(OWLOntologyStorageException ex){
             logger.error("OntologyStorageException",Logging.Origin.USAGE,ex);
             return false;}
        return true;}
   
    /** creates a new IRI for the given name in the extension ontology.
     * 
     * @param name the name of the class.
     * @return the new IRI, or null if there is no extension ontology.
     */
    IRI createItemIRI(String name) {
        assert(extensionOntology != null) : "The extension ontology must be defined";
        return IRI.create(extensionOntologyIRI +"#"+name);}

    
    /** creates the change axioms for removing the individual from all ontologies.*/
    public void removeIndividual(OWLNamedIndividual individual) {
        synchronized(changeAxioms) {
            logger.info("OWLIndividual removed", individual);
            applyChanges(true);
            individual.accept(remover);
            changeAxioms.addAll(remover.getChanges());
            remover.reset();}}
    
    /** creates the change axioms for removing all individuals in the given class. */
    public void removeAllIndividuals(OWLClass ocl) {
        synchronized(changeAxioms) {
            logger.info("OWLIndividuals removed from class", ocl);
            applyChanges(true);
            NodeSet<OWLNamedIndividual> individuals = null;
            try{individuals = reasoner.getInstances(ocl, false);}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, ocl, ex);
                return;}
            for (OWLNamedIndividual ind : individuals.getFlattened()) ind.accept(remover);
            changeAxioms.addAll(remover.getChanges());
            remover.reset();}}
    
    //--------------------- Accessing Components of the Ontology ---------------//
    
    /** returns the first OWLDataProperty in the extensionOntology's import closure 
     * whose fragment equals 'name', or null if there is no such property.
     * <p>To this end it searches through the imports closure of the extension ontology and the first one
     * it finds with the given name is returned. 
     */
    public OWLDataProperty getFirstOWLDataProperty(String name) {
        for(OWLDataProperty property : extensionOntology.getDataPropertiesInSignature(true) ) {
              if(property.getIRI().getFragment().equals(name)) return property;}
        return null;}
    
    /** returns the first OWLObjectProperty in the extensionOntology's import closure 
     * whose fragment equals 'name', or null if there is no such property.
     * <p>To this end it searches through the imports closure of the extension ontology and the first one
     * it finds with the given name is returned. */
    public OWLObjectProperty getFirstOWLObjectProperty(String name) {
        for(OWLObjectProperty property : extensionOntology.getObjectPropertiesInSignature(true) ) {
              if(property.getIRI().getFragment().equals(name)) return property;}
        return null;}
   
    /** returns a set of data properties with the given name.
     * <p>
     * There may be more than data property with the given name in the imports closure.
     * 
     * @param name the name of an OWL data property.
     * @param background if true then the background ontology is searched, otherwise the extension ontology.
     * @return the set of OWLDataProperties with the given name. 
     */
    Set<OWLDataProperty> getOWLDataProperties(String name, boolean background) {
        synchronized(changeAxioms) {
            applyChanges(false);
            OWLOntology ontology = null;
            if(background && backgroundOntology != null) ontology = backgroundOntology;
            if(!background && extensionOntology != null) ontology = extensionOntology;
            assert (ontology != null) : "No ontology is defined yet";
            Set<OWLDataProperty> properties = new HashSet<OWLDataProperty>();
            for(OWLOntology ont : ontology.getImportsClosure()) {
                IRI iri = IRI.create(ont.getOntologyID().getDefaultDocumentIRI().toString()+"#"+name);
                for(OWLEntity entity : ont.getEntitiesInSignature(iri, false)) {
                        if(entity.isOWLDataProperty()) properties.add(entity.asOWLDataProperty());}}
            return properties;}}
    
    /** returns a set of object properties with the given name.
     * <p>
     * There may be more than one object property with the given name in the imports closure.
     * 
     * @param name the name of an OWL named individual.
     * @param background if true then the background ontology is searched, otherwise the extension ontology.
     * @return the set of OWLObjectProperties with the given name. 
     */
    Set<OWLObjectProperty> getOWLObjectProperties(String name, boolean background) {
        synchronized(changeAxioms) {
            applyChanges(false);
            OWLOntology ontology = null;
            if(background && backgroundOntology != null) ontology = backgroundOntology;
            if(!background && extensionOntology != null) ontology = extensionOntology;
            assert (ontology != null) : "No ontology is defined yet";
            Set<OWLObjectProperty> properties = new HashSet<OWLObjectProperty>();
            for(OWLOntology ont : ontology.getImportsClosure()) {
                IRI iri = IRI.create(ont.getOntologyID().getDefaultDocumentIRI().toString()+"#"+name);
                for(OWLEntity entity : ont.getEntitiesInSignature(iri, false)) {
                    if(entity.isOWLObjectProperty()) properties.add(entity.asOWLObjectProperty());}}
            return properties;}}
     
    /** returns an OWL Object- or Data-Property, depending on the isDataProperty flag.
     * 
     * @param name the name of the property.
     * @param isDataProperty indicated whether this is a data- or object property.
     * @return the OWL property.
     * <p>
     * If there is already a property with this name in the background ontology or the
     * extension ontology then this property is returned.
     * To this end it searches for such a property in the imports closure of the extension ontology.
     * The first one it finds is returned.
     * 
     */
    OWLProperty makeOWLProperty(String name, boolean isDataProperty) { 
        OWLOntology ontology = extensionOntology;
        if(ontology == null) ontology = backgroundOntology;
        assert(ontology != null): "No ontology defined yet.";
        for(OWLProperty property : 
                isDataProperty ? ontology.getDataPropertiesInSignature(true) : 
                                 ontology.getObjectPropertiesInSignature(true)) {
              if(property.getIRI().getFragment().equals(name)) return property;}
        IRI iri = createItemIRI(name);
        return isDataProperty ? dataFactory.getOWLDataProperty(iri) : dataFactory.getOWLObjectProperty(iri);}
    
    
    /** yields all OWLDataProperties in the extension ontology whose fragment starts with the given prefix. */
    ArrayList<OWLDataProperty> findOWLDataProperties(String prefix) {
        ArrayList<OWLDataProperty> properties = new ArrayList<OWLDataProperty>();
        for(OWLDataProperty property : extensionOntology.getDataPropertiesInSignature(true)) {
              if(property.getIRI().getFragment().startsWith(prefix)) properties.add(property);}
        return properties;}
    
     /** yields all OWLObjectProperties in the extension ontology whose fragment starts with the given prefix. */
    ArrayList<OWLObjectProperty> findOWLObjectProperties(String prefix) {
        ArrayList<OWLObjectProperty> properties = new ArrayList<OWLObjectProperty>();
        for(OWLObjectProperty property : extensionOntology.getObjectPropertiesInSignature(true)) {
              if(property.getIRI().getFragment().startsWith(prefix)) properties.add(property);}
        return properties;}
    
    /** returns true if the property is from the background ontology */
    boolean propertyIsFromBackground(OWLProperty property) {
        if(backgroundOntology == null) return false;
        assert(extensionOntology != null): "No extension ntology defined yet.";
        return !property.getIRI().getStart().equals(extensionOntology.getOntologyID().getDefaultDocumentIRI()+"#");}
    
    /** returns a set of OWLClasses with the given name.
     * 
     * <br/>
     * There may be more than one class with the given name in the imports closure.
     * 
     * @param name an OWLClass name.
     * @return the set of OWLClasses with the given name. 
     */
    public ArrayList<OWLClass> getOWLClasses(String name) {
        synchronized(changeAxioms) {
            applyChanges(false);
            OWLOntology ontology = (extensionOntology != null) ?  extensionOntology : backgroundOntology;
            assert (ontology != null) : "No ontology is defined yet";
            ArrayList<OWLClass> classes = new ArrayList<OWLClass>();
            for(OWLOntology ont : ontology.getImportsClosure()) {
                IRI iri = IRI.create(ont.getOntologyID().getDefaultDocumentIRI().toString()+"#"+name);
                for(OWLEntity entity : ont.getEntitiesInSignature(iri, false)) {
                    if(entity.isOWLClass()) classes.add(entity.asOWLClass());}}
            return classes;}}
    
    /** returns a set of OWLNamedIndividuals with the given name.
     * <p>
     * There may be more than one individual with the given name in the imports closure.
     * 
     * @param name the name of an OWL named individual.
     * @param background if true then the background ontology is searched, otherwise the extension ontology.
     * @return the set of OWLNamedIndividuals with the given name.
     */
    Set<OWLNamedIndividual> getOWLIndividuals(String name, boolean background) {
        synchronized(changeAxioms) {
            applyChanges(false);
            OWLOntology ontology = null;
            if(background && backgroundOntology != null) ontology = backgroundOntology;
            if(!background && extensionOntology != null) ontology = extensionOntology;
            assert (ontology != null) : "No ontology is defined yet";
            Set<OWLNamedIndividual> individuals = new HashSet<OWLNamedIndividual>();
            for(OWLOntology ont : ontology.getImportsClosure()) {
                    IRI iri = IRI.create(ont.getOntologyID().getDefaultDocumentIRI().toString()+"#"+name);
                for(OWLEntity entity : ont.getEntitiesInSignature(iri, false)) {
                    if(entity.isOWLNamedIndividual()) individuals.add(entity.asOWLNamedIndividual());}}
            return individuals;}}
    
    //--------------------- END: OWL Ontology Management -----------------------//
    
    //------------------------ BEGIN: OWL Reasoner -----------------------------//
    
    
    /** the currently installed reasoner. */
    private OWLReasoner reasoner = null;
    
    /** the name of the reasoner. */
    private String reasonerName = "HermiT"; 
    
    /** a parameter for the reasoner. */
    private boolean buffering = true;
    
    /** a parameter for the reasoner. */
    private boolean monitorProgress = false;
    
    /** a parameter for the reasoner */
    private int timeout = 10000;
    
    /** returns the names of all available reasoners. */
    public static Set<String> getReasoners() {
       Set<String> reasoners = new HashSet<String>();
       reasoners.add("HermiT");
       reasoners.add("Pellet");
       reasoners.add("FaCT++");
       return reasoners;}

    /** installs a reasoner for the extension ontology. 
     * 
     * @param reasonerName the name of the reasoner.
     * @param buffering  if true, a buffering reasoner is to be installed.
     * @param monitorProgress if true,a progress monitor is to be installed. 
     * @param timeout the timeout (in ms) for complex reasoning tasks. 
     * @return the installed reasoner, or null if there is not yet an extension ontology.
    * 
    * <p>If there is already an extension ontology then the reasoner is loaded. 
    * Otherwise the parameters are stored and the reasoner is loaded after the
    * extension ontology is loaded.
     */ 
    public OWLReasoner setReasoner(String reasonerName, boolean buffering, boolean monitorProgress, int timeout) {
       this.reasonerName = reasonerName;
       this.buffering = buffering;
       this.monitorProgress = monitorProgress;
       this.timeout = timeout;
       if(extensionOntology != null) return loadReasoner();
       return null;}
   
    /** installs a reasoner for the extension ontology. 
     * 
     * @return the installed reasoner or null if something goes wrong. 
     */ 
    private OWLReasoner loadReasoner() {
        if(reasonerName == null) return null;
        if(reasoner != null) {
           logger.info("Reasoner unloaded",reasoner);
           reasoner.dispose();
           reasoner = null;} 
        ReasonerProgressMonitor progressMonitor = (monitorProgress ? new ConsoleProgressMonitor() : new NullReasonerProgressMonitor());
        OWLReasonerConfiguration config = new SimpleConfiguration(progressMonitor,timeout);
        
        if(reasonerName.equals("HermiT")) {
            OWLReasonerFactory reasonerFactory = new Reasoner.ReasonerFactory();
            reasoner = buffering ? reasonerFactory.createReasoner(extensionOntology,config) :
                    reasonerFactory.createNonBufferingReasoner(extensionOntology,config);
        }
        if(reasonerName.equals("Pellet")) {
            PelletReasonerFactory reasonerFactory = new PelletReasonerFactory();
            reasoner = buffering ? reasonerFactory.createReasoner(extensionOntology,config) :
                    reasonerFactory.createNonBufferingReasoner(extensionOntology,config);
           
        }
           if(reasonerName.equals("FaCT++")) {
            FaCTPlusPlusReasonerFactory reasonerFactory = new FaCTPlusPlusReasonerFactory();
            reasoner = buffering ? reasonerFactory.createReasoner(extensionOntology,config) :
                    reasonerFactory.createNonBufferingReasoner(extensionOntology,config);
        }
        
        if(reasoner != null) {logger.info("Reasoner loaded", reasoner);}
        return reasoner;}
   
    //-------------------------- END: OWL Reasoner -----------------------------//
    
    // ------------- BEGIN: Queries to the Reasoner --------------------------//
    
    /** returns for the given OWL class the direct/transitive superclasses of null if an error occured. */
    Set<OWLClass> getSuperclasses(OWLClass ocl, boolean direct) {
        synchronized(changeAxioms) {
            applyChanges(true);
            reasoner.flush();
            Set<OWLClass> superclasses = null;
            try {superclasses = reasoner.getSuperClasses(ocl, direct).getFlattened();}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, ocl, ex);}
            return superclasses;}}

   
    /** returns the set of lowest OWL-Classes containing the individual, or null if an error occurred.
    @param individual is the OWL-individual*/
    public Set<OWLClass> getOWLClasses(OWLNamedIndividual individual) {
       synchronized(changeAxioms) {
           applyChanges(true);
           reasoner.flush();
           Set<OWLClass> owlclasses = null;
           try {owlclasses = reasoner.getTypes(individual,true).getFlattened();}
           catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, individual, ex);}
           return owlclasses;}}
   
    
    /** returns all OWL-instances for the class expression. */
    public Set<OWLNamedIndividual> getInstances(OWLClassExpression cle) {
        synchronized(changeAxioms) {
            applyChanges(true);
            reasoner.flush();
            Set<OWLNamedIndividual> individuals = null;
            try{individuals = reasoner.getInstances(cle, false).getFlattened();}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, cle, ex);}
            return individuals;}}
    
    /** returns the OWL-instances for the class expression, or null if an error occurred.
     * 
     * @param cles an OWL class-expression in Manchester Syntax
     * @return the set of OWL-individuals which are instances of the class.
     * @throws ParserException if the expression could not be parsed.
     */
    public Set<OWLNamedIndividual> getInstances(String cles) throws ParserException {
        OWLClassExpression cle = parseClassExpression(cles);
        synchronized(changeAxioms) {
            applyChanges(true);
            reasoner.flush();
            NodeSet<OWLNamedIndividual> indnode = null;
            try{indnode = reasoner.getInstances(cle, false);}
            catch(Throwable ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, cle, ex);}
            if(indnode != null) return indnode.getFlattened();
            return null;}}
    
    /** parses a class expression string to obtain a class expression.
     * 
     * @param classExpressionString The class expression string
     * @return The corresponding class expression
     * @throws ParserException if the class expression string is malformed or contains unknown entity names.
     */
    public OWLClassExpression parseClassExpression(String classExpressionString) throws ParserException {
        ManchesterOWLSyntaxEditorParser parser = new ManchesterOWLSyntaxEditorParser(dataFactory, classExpressionString);
        parser.setDefaultOntology(extensionOntology);
        parser.setOWLEntityChecker(entityChecker);
        return parser.parseClassExpression();}
    
       
    /** returns true if the extension ontology is consistent.
     * <p>The reasoner may throw a ReasonerInterruptedException or a TimeOutException.
     * In this case an error message is logged and the method returns false.*/
    public boolean isConsistent() {
        synchronized(changeAxioms) {
            applyChanges(false);
            reasoner.flush();
            boolean consistent = false;
            try{consistent = reasoner.isConsistent();}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, ex);}
            return consistent;}}
    
    
    // ---------------- END: Queries to the Reasoner --------------------------//
    
   
    
    /** turns a Java object into an OWLLiteral, if possible. 
     * 
     * @param object a Java object which can be turned into an OWL newLiterals.
     * @return the OWL literal, or null if the mapping is not possible. 
     */
    static OWLLiteral object2Literal(Object object) {
        if(object instanceof String) return dataFactory.getOWLLiteral((String)object);
        if(object instanceof Number) { 
            if(object instanceof Byte)      return dataFactory.getOWLLiteral(object.toString(),OWL2Datatype.XSD_BYTE);
            if(object instanceof Short)     return dataFactory.getOWLLiteral(object.toString(),OWL2Datatype.XSD_SHORT);
            if(object instanceof Integer)   return dataFactory.getOWLLiteral((Integer)object);
            if(object instanceof Long)      return dataFactory.getOWLLiteral(object.toString(),OWL2Datatype.XSD_LONG);
            if(object instanceof Float)     return dataFactory.getOWLLiteral((Float)object);
            if(object instanceof Double)    return dataFactory.getOWLLiteral((Double)object); }
        if(object instanceof Boolean)       return dataFactory.getOWLLiteral((Boolean)object);
        return null;
    }
    
    
     /** turns Java objects into  OWLLiterals, if possible. 
     * 
     * @param object a Java object, an array or a collection of objects which can be turned into an OWL newLiterals.
     * @param literals a collection to put the literals into.
     * @return true if all conversions succeeded.
     */
    boolean  objects2Literals(Object object, final Collection<OWLLiteral> literals) {
        return Utilities.map(object, new Function() {
            @Override
            public boolean apply(Object value) {
                OWLLiteral literal = object2Literal(value);
                if(literal == null) {
                    logger.error("Unmappable data",Logging.Origin.JAVA,value);
                    return false;}
                else literals.add(literal);
                return true;}});}
    
     
    
   
    
    
    /** parses OWL literals into Java Objects. 
     * 
     * <p>
     * An error message is logged if the parsing failed. 
     * @param literal an OWL-Literals.
     * @return the parsed Java-object, or null if the parsing failed. 
     */
    Object literal2Object(OWLLiteral literal) {
        try {
            OWL2Datatype type = literal.getDatatype().getBuiltInDatatype();
        switch(type) {
            case OWL_RATIONAL: {
                String[] s = literal.getLiteral().split("/");
                int n = Integer.parseInt(s[0]);
                int m = 1;
                if(s.length > 1) m =  Integer.parseInt(s[1]);
                return new Double((double)n/(double)m);}
            case OWL_REAL:          return literal.parseDouble();
            case RDF_PLAIN_LITERAL: return literal.getLiteral();
            case RDF_XML_LITERAL:   return literal.getLiteral();
            case XSD_ANY_URI:       return literal.getLiteral();
            case XSD_BASE_64_BINARY: return literal.parseInteger();
            case XSD_BOOLEAN:       return literal.parseBoolean();
            case XSD_BYTE:          return  Byte.parseByte(literal.getLiteral());
            case XSD_DATE_TIME:     return literal.getLiteral();
            case XSD_DATE_TIME_STAMP: return literal.getLiteral();
            case XSD_DECIMAL:       return literal.parseInteger();
            case XSD_DOUBLE:        return literal.parseDouble();
            case XSD_FLOAT:         return literal.parseFloat();
            case XSD_HEX_BINARY:    return literal.parseInteger(); 
            case XSD_INT:           return literal.parseInteger(); 
            case XSD_INTEGER:       return literal.parseInteger(); 
            case XSD_LANGUAGE:      return literal.getLiteral();
            case XSD_LONG:          return Long.parseLong(literal.getLiteral()); 
            case XSD_NAME:          return literal.getLiteral();
            case XSD_NCNAME:        return literal.getLiteral();
            case XSD_NEGATIVE_INTEGER: return literal.parseInteger(); 
            case XSD_NMTOKEN:       return literal.getLiteral();
            case XSD_NON_NEGATIVE_INTEGER:  return literal.parseInteger(); 
            case XSD_NON_POSITIVE_INTEGER:  return literal.parseInteger(); 
            case XSD_NORMALIZED_STRING:     return literal.getLiteral();
            case XSD_POSITIVE_INTEGER:      return literal.parseInteger(); 
            case XSD_SHORT:         return (short)literal.parseInteger(); 
            case XSD_STRING:        return literal.getLiteral();
            case XSD_TOKEN:         return literal.getLiteral();
            case XSD_UNSIGNED_BYTE: return (byte)literal.parseInteger(); 
            case XSD_UNSIGNED_INT:  return literal.parseInteger(); 
            case XSD_UNSIGNED_LONG: return Long.parseLong(literal.getLiteral()); 
            case XSD_UNSIGNED_SHORT: return Short.parseShort(literal.getLiteral());
            default: logger.error("Unknown OWL-literal type",Logging.Origin.REASONER,type,literal);}}
        catch(NumberFormatException ex) {
            logger.error("Literal can not be parsed to OWL",Logging.Origin.REASONER,literal);
            return null;
        }
        return null;
    }
 
    
    //-------------------------- BEGIN: OWL Axioms -----------------------------//
    
      
    /** stores axioms for enforcing changes to the ontology. */
    private final ArrayList<OWLOntologyChange> changeAxioms = new ArrayList<OWLOntologyChange>();
        
    /** stores information about the EquivalentClass annotation. */
    private final ArrayList<Equivalence> equivalences = new ArrayList<Equivalence>();
    
    /** adds the given axioms as addAxioms to the change Axioms. 
    * No change is made to the ontology itself.
    */
    void addOWLAxioms(List<OWLAxiom> axioms) {
       synchronized(changeAxioms) {
           for(OWLAxiom axiom : axioms) {
               changeAxioms.add(new AddAxiom(extensionOntology,axiom));
               logger.info("OWL-axiom added", axiom);}
           if(debugMode) applyChanges(true);}}
   
    /** adds the given axiom as addAxiom to the change Axioms. 
    * No change is made to the ontology itself.
    */
    void addOWLAxiom(OWLAxiom axiom) {
       synchronized(changeAxioms) {
           changeAxioms.add(new AddAxiom(extensionOntology,axiom)); 
           logger.info("OWL-axiom added", axiom);
           if(debugMode) applyChanges(true);}}
   
    /** adds the given axiom as removeAxiom to the change Axioms. 
    * No change is made to the ontology itself.
    */
    void removeOWLAxiom(OWLAxiom axiom) {
        synchronized(changeAxioms) {
           changeAxioms.add(new RemoveAxiom(extensionOntology,axiom)); 
           logger.info("OWL-axiom removed", axiom);
           if(debugMode) applyChanges(true);}}
   
     /** creates an object-property assertion axiom.
     * 
     * @param individual the OWL-individual whose property is to be axiomatised.
     * @param property   the object-property.
     * @param value      the filler.
     * @return           the assertion-axiom. 
     */
    static OWLAxiom getObjectPropertyAxiom(OWLIndividual individual, OWLObjectProperty property, OWLIndividual value) {
        return dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, value);}
    
    /** applies the change axioms to the extension ontology.
    * <p>If the system is in debug mode all changes are flushed to the reasoner and the 
    * consistency  is checked.
    * 
    * @param  checkConsistency causes consistency check in the debug mode.
    * @return true if the operation succeeded and the classes are consistent (if it is checked).
     * 
     * <p>This method is called either in the compilation phase where no thread safety is implemented,
     * or from methods which are protected by changeAxioms as monitor.
    */
    public boolean applyChanges(boolean checkConsistency) {
        if(changeAxioms.isEmpty()) return true;
        logger.entering(changeAxioms);
        if(!changeAxioms.isEmpty()) {
            ontologyManager.applyChanges(changeAxioms);
            changeAxioms.clear();}
        
        boolean consistent = false;
        if(debugMode && checkConsistency && reasoner != null) {
            try {
                reasoner.flush();
                consistent = reasoner.isConsistent();}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER);}}
        
        logger.exiting();
        return consistent;}
   
    /** turns the equivalences into OWL axioms. */
    void addEquivalences() {
        for(Equivalence equiv : equivalences) equiv.addEquivalentClassesAxiom();
            equivalences.clear();}
    
    
    //-------------------------- END: OWL Axioms -------------------------------//
    
    // ----------------------BEGIN: Java <-> OWL --------------------------------//
    
        
    /** stores counters for giving translated Java Objects individual names. */
    private final Map<OWLClass,Integer>  nameCounter = new HashMap<OWLClass,Integer>();
    
    /** returns a counter for the OWL class. 
     * 
     * @param ocl an OWL class
     * @return a counter
     * <p> The counter is automatically incremented.
     */
    private int getNameCounter(OWLClass ocl) {
        synchronized(nameCounter) {
            Integer counter = nameCounter.get(ocl);
            if(counter == null) counter = 0;
            nameCounter.put(ocl,counter+1);
            return counter;}}
    
    /** turns the annotated Java class into an OWL class. 
     * <p>
     * The subclass-superclass relations at the Java side are transferred to the OWL-side.
     * This method assumes that annotated superclasses and interfaces of jcl are already mapped to OWL-classes 
     * The annotation OWLSuperClasses also causes the addition of superclass axioms. 
     * 
     * @param jcl a Java class or interface.
     * @return a new ClassWrapper with the correspondence between Java class and OWL class.
     */
    ClassWrapper class2OWL(Class jcl) {
        assert (extensionOntology != null) : "The extension ontology is not defined";
        logger.entering(jcl);
        
        J2OWLClass j2owlannotation = (J2OWLClass)jcl.getAnnotation(J2OWLClass.class);
        
        J2OClassManager classManager = manager.getClassManager();
        String oclName = j2owlannotation.name();      
        if(oclName.isEmpty()) oclName = jcl.getSimpleName(); 
        OWLClass ocl = dataFactory.getOWLClass(createItemIRI(oclName)); // a new OWLClass
        logger.event(Logging.Semantics.OCC, "New OWL class created ", ocl);
        addOWLAxiom(dataFactory.getOWLDeclarationAxiom(ocl));
        addClassAttribute(ocl,jcl);
        OWLClass superocl = classManager.getOWLClass(jcl.getSuperclass(), false, false);
        
        String equivExpression = j2owlannotation.EquivalentClass();
        if(equivExpression.isEmpty()){
            if(superocl != null) {addSuperclassAxiom(ocl,superocl);}
            for(String owlSC : j2owlannotation.OWLSuperClasses().split(",")) {
                for(OWLClass socl : getOWLClasses(owlSC.trim())) {
                    addSuperclassAxiom(ocl,socl);}}}
        else {
            Set<OWLClassExpression> superclasses = new HashSet<OWLClassExpression>();
            if(superocl != null) superclasses.add(superocl);
            for(String owlSC : j2owlannotation.OWLSuperClasses().split(",")) {
                for(OWLClass socl : getOWLClasses(owlSC.trim())) {
                    superclasses.add(socl);}}
            equivalences.add(new Equivalence(jcl,ocl,equivExpression,superclasses));}
        
        logger.exiting(jcl,ocl);
        return classManager.addClassCorrespondence(jcl,ocl);}
   
    /** add the subclass-superclass axiom to the extension ontology. */
    void addSuperclassAxiom (OWLClass subclass, OWLClass superclass) {
        addOWLAxiom(dataFactory.getOWLSubClassOfAxiom(subclass, superclass));}
   
    /** add the class-equivalent class axiom to the extension ontology. */
    void addEquivalentClassesAxiom(OWLClass subclass, Set<OWLClassExpression> superclasses) {
       OWLClassExpression clex = dataFactory.getOWLObjectIntersectionOf(superclasses);
       addOWLAxiom(dataFactory.getOWLEquivalentClassesAxiom(subclass,clex));}
   
    /** adds the OWL-attribute to the given OWL-class which expresses that this class is generated from 
    * the given Java class.
    * 
    * @param ocl an OWL-class.
    * @param jcl a Java-class.
    */
    void addClassAttribute(OWLClass ocl, Class jcl) {
       OWLAnnotationSubject subject = ocl.getIRI();
       OWLAnnotationValue value = dataFactory.getOWLLiteral(jcl.getName());
       addOWLAxiom(dataFactory.getOWLAnnotationAssertionAxiom(classAnnotationProperty, subject, value));
   }
   
        
    /** adds the OWL-attribute to the given OWL-property which expresses that this property is generated from 
    * the given Java method.
    * 
    * @param property an OWL property.
    * @param jcl a Java-class.
    * @param method a Java method.
    */
    void addMethodAttribute(OWLProperty property, Class jcl, Method method) {
       OWLAnnotationSubject subject = property.getIRI();
       OWLAnnotationValue value = dataFactory.getOWLLiteral(jcl.getName() + "."+method.getName());
       addOWLAxiom(dataFactory.getOWLAnnotationAssertionAxiom(methodAnnotationProperty, subject, value));}
  
    /** adds the Java class names as OWL-attribute to the extension ontology. 
    * 
    * @param classes a collection of Java classes.
    */
    void addJavaClassesAsAnnotation(Collection<Class> classes) {
       if(classes.isEmpty()) return;
       String classnames = "";
       for(Class jcl : classes) {if(jcl.getEnclosingClass() == null) classnames += jcl.getName() + ",";}
       OWLAnnotationSubject subject = extensionOntology.getOntologyID().getDefaultDocumentIRI();
       OWLAnnotationValue value = dataFactory.getOWLLiteral(classnames.substring(0,classnames.length()-1));
       addOWLAxiom(dataFactory.getOWLAnnotationAssertionAxiom(classesAnnotationProperty, subject, value));}
   
    /** retrieves the Java class names from the OWL-attribute of the extension ontology. */
    String[] getJavaClassesFromAnnotation() {
       for(OWLAnnotation annotation : extensionOntology.getAnnotations()) {
           if(annotation.getProperty().equals(classesAnnotationProperty)) {
               return ((OWLLiteral)(annotation.getValue())).getLiteral().split(",");}}
       return new String[]{};}
    
    
    /** returns the axiom expressing individual membership in the OWL-class. */ 
    static OWLAxiom getMembershipAxiom(OWLIndividual individual, OWLClass ocl) {
          return dataFactory.getOWLClassAssertionAxiom(ocl, individual);}
      
     
    /** turns the value into property assertion axioms for the individual.
     * <p>
     * If value is an array or a collection type all elements are turned into role-fillers.
     * 
     * @param individual    an OWL-individual.
     * @param property      an OWLDataProperty.
     * @param value         a Java-object which has to be turned into OWL role-fillers,
     * @param axioms        the list where the axioms are to be added. 
     * @return              true if the assignment was possible. 
     */
    boolean makeDataPropertyAxioms(final OWLNamedIndividual individual, final OWLDataProperty property, Object value, final ArrayList<OWLAxiom> axioms)  {
        return Utilities.map(value,
                new Function() {
                    @Override
                    public boolean apply(Object value){
                        OWLLiteral literal = object2Literal(value);
                        if(literal == null) {
                            logger.error("Unmappable data",Logging.Origin.JAVA,value);
                            return false;}
                        axioms.add(dataFactory.getOWLDataPropertyAssertionAxiom((OWLDataProperty)property, 
                                individual, literal));
                        return true;}});}
    
     /** turns the value into property assertion axioms for the individual.
     * <p>
     * If value is an array or a collection type all elements are turned into role-fillers.
     * 
     * @param individual    an OWL-individual.
     * @param property      an OWLObjectProperty.
     * @param value         a Java-object which has to be turned into OWL role-fillers,
     * @param axioms        the list where the axioms are to be added. 
     * @return              true if the assignment was possible. 
     */
    void makeObjectPropertyAxioms(OWLNamedIndividual individual, final OWLObjectProperty property, Object value, final ArrayList<OWLAxiom> axioms)  {
        for(OWLNamedIndividual otherIndividual : manager.getIndvidualManager().makeIndividuals(value)) {
            axioms.add(dataFactory.getOWLObjectPropertyAssertionAxiom(property,individual, otherIndividual));}}
    
    
    /** Creates for the extension ontology a new OWL individual for a given Java Object. 
    * <p>
    * The name of the new individual is determined as follows:
    * <ul>
    * <li> If object's class has a J2OWLClass attribute with a naming()-parameter, and this parameter is 
    *   not the empty string, then this method is tried. The default is "getName().</li>
    * <li>If this does not work then the name is <OWLClass name>_<counter> </li>
    * </ol>
    * <p> No data or object properties are yet transferred to the individual.
    * 
    * @param object The object to be mapped to an OWL individual.
    * @param ocl an OWL-class to be used for determining a name for the individual.
    * @return a new OWL individual.
    */
    OWLNamedIndividual createOWLIndividual(Object object, OWLClass ocl) {
        assert(extensionOntology != null) : "The extension ontoloy must be defined";
        Class jcl = object.getClass();
        String indName = null;
        Method getName = null;
        String naming = "getName";
        J2OWLClass j2owlannotation = (J2OWLClass)(object.getClass().getAnnotation(J2OWLClass.class));
        if(j2owlannotation != null && !j2owlannotation.naming().isEmpty()) 
            naming = j2owlannotation.naming();
        try {getName = jcl.getMethod(naming);} catch (Exception ex) {}
        if(getName != null) {
            try {indName = ((String)(getName.invoke(object))).replace(' ', '_');}  
            catch (Exception ex) {}}
        if(indName == null) {
            indName = ocl.getIRI().getFragment()+"_"+getNameCounter(ocl);}
        OWLNamedIndividual individual = dataFactory.getOWLNamedIndividual(IRI.create(extensionOntologyIRI+"#"+indName));
        return individual;
    } 
    
      /** inserts a new OWL-individual with the given name into the OWL-class.
     * @param name the name of the individual
     * @param ocl  the OWL-class
     * @return the new OWL-individual.
     */
    OWLNamedIndividual insertOWLIndividual(String name, OWLClass ocl) {
        OWLNamedIndividual individual = dataFactory.getOWLNamedIndividual(IRI.create(extensionOntologyIRI+"#"+name));
        synchronized(changeAxioms) {
            addOWLAxiom(dataFactory.getOWLDeclarationAxiom(individual));
            addOWLAxiom(getMembershipAxiom(individual,ocl));}
        return individual;}
    
     // ----------------------END: Java <-> OWL --------------------------------//
    

  
    // --------------BEGIN: Changing Property Values --------------------------//
  
    
    /** inserts/removes a data property.
     * 
     * @param individual the individual for which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param literal    the filler to be inserted or removed. 
     * @param insert     if true then the value is inserted, otherwise removed.
     */
    private void changeDataProperty(OWLIndividual individual, OWLDataProperty property, OWLLiteral literal, boolean insert) {
        if(literal == null) return;
        OWLAxiom axiom =  dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, literal);
        if(insert) addOWLAxiom(axiom);
        else removeOWLAxiom(axiom);}
    
    /** maps a data-property change caused by an add-method or a remove-method to the ontology.
     * 
     * @param individual the individual for which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param value      the filler to be inserted or removed. 
     * @param insert     if true then the value is inserted, otherwise removed.
     * @return           true if the change was possible. 
     */
    boolean changeDataProperty(final OWLIndividual individual, final OWLDataProperty property, Object value, final boolean insert) {
        synchronized(changeAxioms) {
            return Utilities.map(value, new Function() {
                @Override
                public boolean apply(Object value) {
                    OWLLiteral literal = object2Literal(value);
                    if(literal == null) {
                        logger.error("Unmappable data",Logging.Origin.JAVA,value);
                        return false;}
                    else changeDataProperty(individual,property,literal,insert);
                    return true;}});}}
    
    
    
    
    
    
    
    /** maps a data-property change caused by a set-method to the ontology.
     * 
     * @param individual the individual form which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param oldValue   the old filler.
     * @param newValue   the new filler.
     * @return           true if the change was possible. 
     */
    boolean changeDataProperty(OWLIndividual individual, OWLDataProperty property, Object oldValue, Object newValue) {
        synchronized(changeAxioms) {
            return  changeDataProperty(individual,property,oldValue,false) &&
                    changeDataProperty(individual,property,newValue,true);}}
        
    /** maps an object-property change cause by an add-method or a remove-method to the ontology.
     * 
     * @param individual the individual for which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param value      the filler to be inserted or removed. 
     * @param insert     if true then the value is inserted, otherwise removed.
     */
    void changeObjectProperty(OWLIndividual individual, OWLObjectProperty property, OWLIndividual value, boolean insert) {
          OWLAxiom axiom =  dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, value); 
          if(insert) addOWLAxiom(axiom);
          else removeOWLAxiom(axiom);}
    
    /** maps an object-property change cause by a set-method to the ontology.
     * 
     * @param individual the individual for which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param oldValue   the old filler.
     * @param newValue   the new filler.
     */
    void changeObjectProperty(OWLNamedIndividual individual, OWLObjectProperty property, OWLIndividual oldValue, OWLIndividual newValue) {
        synchronized(changeAxioms) {
            removeOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, oldValue));
            addOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, newValue));}}
  
    /** sets the object as new data property of the individual.
     * <p>
     * All previous values are removed and the new value is made the only property value.
     * If the value cannot be mapped to OWL then an error is logged and the method returns false.
     * <p>
     * Only the change axioms are created. 
     * 
     * @param individual an OWLNamedIndividual.
     * @param property an OWLDataProperty.
     * @param value a Java object which gan be mapped to an OWLLiteral.
     * @return true if the operation succeeded.
     */
    boolean setDataProperty(OWLNamedIndividual individual, OWLDataProperty property, Object value) {
        OWLLiteral literal = object2Literal(value);
        if(literal == null) {
            logger.error("Unmappable data",Logging.Origin.JAVA,value);
            return false;}
        boolean found = false;
        synchronized(changeAxioms) {
            for (OWLLiteral oldLiteral : getDataPropertyLiterals(individual,property)) {
                if(oldLiteral.equals(literal)) {found = true;}
                else removeOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, oldLiteral));} 
            if(!found) {
                addOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, literal));}
        return true;}}
    
    /** sets the values as new data property of the individual.
     * <p>
     * All previous values are removed and the new values are made the only property values.
     * If the values cannot be mapped to OWL then an error is logged and the method returns false.
     * <p>
     * Only the change axioms are created. 
     * 
     * @param individual an OWLNamedIndividual.
     * @param property an OWLDataProperty.
     * @param values a collection of Java object which gan be mapped to an OWLLiteral.
     * @return true if the operation succeeded.
     */
    boolean setDataProperties(OWLNamedIndividual individual, OWLDataProperty property, Collection values) {
        boolean success = true;
        Set<OWLLiteral> newLiterals = new HashSet<OWLLiteral>();
        for(Object value : values) {
            OWLLiteral literal = object2Literal(value);
            if(literal == null) {
                logger.error("Unmappable data",Logging.Origin.JAVA,value);
                success = false;}
            else newLiterals.add(literal);}
        ArrayList<OWLLiteral> oldLiterals = new ArrayList<OWLLiteral>();
        
        synchronized(changeAxioms) {
            for (OWLLiteral oldLiteral : getDataPropertyLiterals(individual,property)) {
                if(newLiterals.contains(oldLiteral)) newLiterals.remove(oldLiteral);
                else oldLiterals.add(oldLiteral);}
        
            for(OWLLiteral oldLiteral : oldLiterals) {
                removeOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, oldLiteral));} 
            for(OWLLiteral newLiteral : newLiterals) {
                addOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, newLiteral));}}
        return success;}
    
    /** sets the values as new data property of the individual.
     * <p>
     * All previous values are removed and the new values are made the only property values.
     * If the values cannot be mapped to OWL then an error is logged and the method returns false.
     * <p>
     * Only the change axioms are created. 
     * 
     * @param individual an OWLNamedIndividual.
     * @param property an OWLDataProperty.
     * @param values  an array of Java objects which gan be mapped to an OWLLiteral.
     * @return true if the operation succeeded.
     */
    boolean setDataProperties(OWLNamedIndividual individual, OWLDataProperty property, Object[] values) {
        boolean success = true;
        Set<OWLLiteral> newLiterals = new HashSet<OWLLiteral>();
        for(Object value : values) {
            OWLLiteral literal = object2Literal(value);
            if(literal == null) {
                logger.error("Unmappable data",Logging.Origin.JAVA,value);
                success = false;}
            else newLiterals.add(literal);}
        ArrayList<OWLLiteral> oldLiterals = new ArrayList<OWLLiteral>();
        
        synchronized(changeAxioms) {
            for (OWLLiteral oldLiteral : getDataPropertyLiterals(individual,property)) {
                if(newLiterals.contains(oldLiteral)) newLiterals.remove(oldLiteral);
                else oldLiterals.add(oldLiteral);}
        
            for(OWLLiteral oldLiteral : oldLiterals) {
                removeOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, oldLiteral));} 
            for(OWLLiteral newLiteral : newLiterals) {
                addOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, newLiteral));}}
        return success;}
    
    private <Property extends OWLProperty> void getMapProperties(String propertyPrefix, Map values,
            boolean isDataProperty,
            Collection<Property> addProperties,
            Collection<Object> addValues,
            Collection<Property> intersectProperties,
            Collection<Object> intersectValues,
            Collection<Property> removeProperties){
        Object[] newKeys = values.keySet().toArray();
        int kSize = newKeys.length;
        String[] newKeyStrings = new String[kSize];
        for(int i = 0; i < kSize; ++i) newKeyStrings[i] = newKeys[i].toString();
        
        int pLength = propertyPrefix.length();
        Set<String> oldKeyStrings = new HashSet<String>();
        
        for(OWLProperty property : isDataProperty? findOWLDataProperties(propertyPrefix): findOWLObjectProperties(propertyPrefix)) {
            String key = property.getIRI().getFragment().substring(pLength);
            oldKeyStrings.add(key);
            for(int i = 0; i < kSize; ++i) {
                if(newKeyStrings[i].equals(key)) {
                    intersectProperties.add((Property)property);
                    intersectValues.add(values.get(newKeys[i]));
                    break;}
            else removeProperties.add((Property)property);}}
            
        for(int i = 0; i < kSize; ++i) {
            if(!oldKeyStrings.contains(newKeyStrings[i])) {
                addProperties.add((Property)makeOWLProperty(propertyPrefix+newKeyStrings[i],isDataProperty));
                addValues.add(values.get(newKeys[i]));}}
    }
    
    
    /** for Map-type attributes: sets the values as new data property of the individual.
     * <p>
     * All previous values are removed and the new values are made the only property values.
     * If the values cannot be mapped to OWL then an error is logged and the method returns false.
     * <p>
     * Only the change axioms are created. 
     * 
     * @param individual an OWLNamedIndividual.
     * @param property an OWLDataProperty.
     * @param values  a map key-> Java objects which gan be mapped to an OWLLiteral.
     * @return true if the operation succeeded.
     */
    boolean setMapDataProperties(OWLNamedIndividual individual, String propertyPrefix, Map values) {
        ArrayList<OWLDataProperty> addProperties = new ArrayList<OWLDataProperty>();
        ArrayList<Object> addValues = new ArrayList<Object>();
        ArrayList<OWLDataProperty> intersectProperties = new ArrayList<OWLDataProperty>();
        ArrayList<Object> intersectValues = new ArrayList<Object>();
        ArrayList<OWLDataProperty> removeProperties = new ArrayList<OWLDataProperty>();
        
        getMapProperties(propertyPrefix,values,true,addProperties,addValues,
                    intersectProperties,intersectValues,removeProperties);
        
        
        for(OWLDataProperty property : removeProperties) {
            for (OWLLiteral oldLiteral : getDataPropertyLiterals(individual,property)) {
                removeOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, oldLiteral));}}
        
        Set<OWLLiteral> newLiterals = new HashSet<OWLLiteral>();
        ArrayList<OWLLiteral> oldLiterals = new ArrayList<OWLLiteral>();
        int length = intersectProperties.size();
        for(int i = 0; i < length; ++i) {
            OWLDataProperty property =  intersectProperties.get(i);
            newLiterals.clear();
            oldLiterals.clear(); 
            if(!objects2Literals(intersectValues.get(i),newLiterals)) return false;
            for (OWLLiteral oldLiteral : getDataPropertyLiterals(individual,property)) {
                if(newLiterals.contains(oldLiteral)) newLiterals.remove(oldLiteral);
                else oldLiterals.add(oldLiteral);}
    
            for(OWLLiteral oldLiteral : oldLiterals) {
                removeOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, oldLiteral));} 
            for(OWLLiteral newLiteral : newLiterals) {
                addOWLAxiom(dataFactory.getOWLDataPropertyAssertionAxiom(property, individual, newLiteral));}}
                 
        ArrayList<OWLAxiom> axioms = new ArrayList<OWLAxiom>();
        length = addProperties.size();
        for(int i = 0; i < length; ++i) 
            makeDataPropertyAxioms(individual,addProperties.get(i),addValues.get(i),axioms);
        addOWLAxioms(axioms);
        return true;}
    
    /** for Map-type attributes: sets the values as new object property of the individual.
     * <p>
     * All previous values are removed and the new values are made the only property values.
     * <p>
     * Only the change axioms are created. 
     * 
     * @param individual an OWLNamedIndividual.
     * @param property an OWLObjectProperty.
     * @param values  a map key-> Java objects which gan be mapped to an OWLNamedIndividual.
     * @return true if the operation succeeded.
     */
    boolean setMapObjectProperties(OWLNamedIndividual individual, String propertyPrefix, Map values) {
        J2OIndividualManager img = manager.getIndvidualManager();
        ArrayList<OWLObjectProperty> addProperties = new ArrayList<OWLObjectProperty>();
        ArrayList<Object> addValues = new ArrayList<Object>();
        ArrayList<OWLObjectProperty> intersectProperties = new ArrayList<OWLObjectProperty>();
        ArrayList<Object> intersectValues = new ArrayList<Object>();
        ArrayList<OWLObjectProperty> removeProperties = new ArrayList<OWLObjectProperty>();
        
        getMapProperties(propertyPrefix,values,false,addProperties,addValues,
                    intersectProperties,intersectValues,removeProperties);
        
        for(OWLObjectProperty property : removeProperties) {
            for (OWLNamedIndividual oldIndividual : getObjectPropertyValues(individual,property)) {
                removeOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, oldIndividual));}}
        
        Set<OWLNamedIndividual> newIndividuals = new HashSet<OWLNamedIndividual>();
        ArrayList<OWLNamedIndividual> oldIndividuals = new ArrayList<OWLNamedIndividual>();
        int length = intersectProperties.size();
        for(int i = 0; i < length; ++i) {
            OWLObjectProperty property =  intersectProperties.get(i);
            newIndividuals.clear();
            oldIndividuals.clear(); 
            newIndividuals = img.makeIndividuals(intersectValues.get(i));
            for (OWLNamedIndividual oldLiteral : getObjectPropertyValues(individual,property)) {
                if(newIndividuals.contains(oldLiteral)) newIndividuals.remove(oldLiteral);
                else oldIndividuals.add(oldLiteral);}
    
            for(OWLNamedIndividual oldIndividual : oldIndividuals) {
                removeOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, oldIndividual));} 
            for(OWLNamedIndividual newIndividual : newIndividuals) {
                addOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, newIndividual));}}
                 
        ArrayList<OWLAxiom> axioms = new ArrayList<OWLAxiom>();
        length = addProperties.size();
        for(int i = 0; i < length; ++i) 
            makeObjectPropertyAxioms(individual,addProperties.get(i),addValues.get(i),axioms);
        addOWLAxioms(axioms);
        return true;}
    
    
    /** adds newIndividual as new role filler for the individual's property */
    void addObjectPropertyValue(OWLNamedIndividual individual, OWLObjectProperty property, OWLNamedIndividual newIndividual) {
         if(newIndividual != null)
            addOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, newIndividual));}
     
    /** removed oldIndividual as new role filler for the individual's property */
    void removeObjectPropertyValue(OWLNamedIndividual individual, OWLObjectProperty property, OWLNamedIndividual oldIndividual) {
         if(oldIndividual != null)
             removeOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, oldIndividual));}
    
    /** sets the newIndividual as the only object property value.
     * <p>
     * All old fillers are removed.
     * 
     * @param individual the individual for which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param newIndividual   the new filler.
     */
    void setObjectProperty(OWLNamedIndividual individual, OWLObjectProperty property, OWLNamedIndividual newIndividual) {
        boolean found = false;
        synchronized(changeAxioms) {
            Set<OWLNamedIndividual> individuals = getObjectPropertyValues(individual,property);
            if(individuals == null) return;// a reasoner error occurred.
            for(OWLNamedIndividual oldIndividual : individuals) {
                if(newIndividual.equals(oldIndividual)) found = true;
                else removeOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, oldIndividual));}
            if(!found)
                    addOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, newIndividual));}}
    
    /** sets the newIndividuals as the only object property value.
     * <p>
     * All old fillers are removed.
     * 
     * @param individual the individual for which the change has to take effect.
     * @param property   the property whose filler is to be changed.
     * @param newIndividuals   the new filler.
     */
    void setObjectProperties(OWLNamedIndividual individual, OWLObjectProperty property, Collection<OWLNamedIndividual> newIndividuals) {
        boolean found = false;
        Set<OWLNamedIndividual> nIndividuals = new HashSet<OWLNamedIndividual>();
        nIndividuals.addAll(newIndividuals);
        ArrayList<OWLNamedIndividual> oIndividuals = new ArrayList<OWLNamedIndividual>();
        synchronized(changeAxioms) {
            Set<OWLNamedIndividual> individuals = getObjectPropertyValues(individual,property);
            if(individuals == null) return; // a reasoner error occurred.
            for(OWLNamedIndividual oldIndividual : individuals) {
                if(newIndividuals.contains(oldIndividual)) nIndividuals.remove(oldIndividual);
                else oIndividuals.add(oldIndividual);}
            for(OWLNamedIndividual oldIndividual : oIndividuals) 
                removeOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, oldIndividual));
            for(OWLNamedIndividual newIndividual : nIndividuals) 
                addOWLAxiom(dataFactory.getOWLObjectPropertyAssertionAxiom(property, individual, newIndividual));}}
     
    /** returns all role-fillers for data-properties as Java objects, or null if an error occurred.
     * <p>
     * OWL-newLiterals which could not be parsed into Java-objects are ignored,
     * only an error message is recorded in the logger.
     * 
     * @param individual the individual whose OWL-role fillers are to he retrieved.
     * @param property  the data-property.
     * @return the set of role-fillers as Java objects, or null if an error occurred.
     */
    synchronized public Set<Object> getDataPropertyValues(OWLNamedIndividual individual, OWLDataProperty property) {
        Set<OWLLiteral> literals = null;
        synchronized(changeAxioms) {
            applyChanges(true);
            reasoner.flush();
            try{literals = reasoner.getDataPropertyValues(individual, property);}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, individual, property, ex);
                return null;}}
        Set<Object> objects = new HashSet<Object>();
        for(OWLLiteral literal : literals) {
            Object value = literal2Object(literal);
            if(value != null) objects.add(value);}
        return objects;}
    
    /** returns all role-fillers for data-properties as OWL-Literals.
     * 
     * @param individual the individual whose OWL-role fillers are to he retrieved.
     * @param property  the data-property.
     * @return the set of role-fillers as OWL-Literals.
     */
    Set<OWLLiteral> getDataPropertyLiterals(OWLNamedIndividual individual, OWLDataProperty property) {
        synchronized(changeAxioms) {
            applyChanges(true);
            reasoner.flush();
            Set<OWLLiteral> literals = null;
            try{literals = reasoner.getDataPropertyValues(individual, property);}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, individual, property, ex);}
            return literals;}}
    
    /** returns all role-fillers for object-properties as OWLNamedindividuals.
     * 
     * @param individual the individual whose OWL-role fillers are to he retrieved.
     * @param property  the object-property.
     * @return the set of role-fillers as OWLNamedindividuals.
     */
    public Set<OWLNamedIndividual> getObjectPropertyValues(OWLNamedIndividual individual, OWLObjectProperty property) {
        synchronized(changeAxioms) {
            applyChanges(true);
            reasoner.flush();
            Set<OWLNamedIndividual> individuals = null;
            try{individuals = reasoner.getObjectPropertyValues(individual, property).getFlattened();}
            catch(OWLRuntimeException ex) {
                logger.error("Reasoner error", Logging.Origin.REASONER, individual, property, ex);}
            return individuals;}}

    
     
    /** synchronises the given object with the corresponding OWL-individual. 
      * <p> This method can be called if the object synchronisation has been turned off for a while.
      * <p> Role fillers (i.a. attribute values of set-valued attributes) which are not yet mapped
      * to OWL are automatically mapped to OWL. 
      * 
      */
    synchronized void synchroniseWithOWL(Object object, OWLNamedIndividual individual) {
        ClassWrapper clw = classManager.getCLW(object.getClass(), false, true);
        synchroniseWithOWLUpwards(object,clw,individual);}
     
    /** synchronises the given object with the corresponding OWL-individual. 
      * <p>The method goes the class hierarchy upwards to synchronise the attributes inherited from superclasses.
      */
    private void synchroniseWithOWLUpwards(Object object, ClassWrapper clw, OWLNamedIndividual individual) {
        J2OIndividualManager img = manager.getIndvidualManager();
         Class jcl = clw.getJavaClass();
         Class superclass = jcl.getSuperclass();
         if(superclass != null) {
             ClassWrapper superclw = classManager.getCLW(superclass, false, true);
             if(superclw != null) synchroniseWithOWLUpwards(object,superclw,individual);}
             
         for(PropertyWrapper pmd : clw.getProperties()) {
             Method getter = pmd.getGetter(jcl);
             Object value = null;
             try {value = getter.invoke(object);}
             catch(Exception ex) {
                 logger.error("Getter error", Logging.Origin.JAVA, object, getter);
                 continue;}
             if(pmd.isDataProperty()) {
                 if(pmd.isMapper())  setMapDataProperties(individual, pmd.getName()+"_", (Map)value);
                 else {
                    OWLDataProperty property = (OWLDataProperty)(pmd.getProperty());
                    if(value instanceof Collection) {
                        setDataProperties(individual, property, (Collection)value);
                        continue;}
                    if(value.getClass().isArray()) {
                        setDataProperties(individual, property, (Object[])value);
                        continue;}
                    setDataProperty(individual, property, value);}}
             else {
                 if(pmd.isMapper()) setMapObjectProperties(individual, pmd.getName()+"_", (Map)value);
                 else {
                    OWLObjectProperty property = (OWLObjectProperty)(pmd.getProperty());
                    if(value instanceof Collection || value.getClass().isArray()) {
                        Set<OWLNamedIndividual> individuals = new HashSet<OWLNamedIndividual>();
                        Object[] mapping = img.getIWRs(value);
                        for(IndividualWrapper vimd : (Set<IndividualWrapper>)(mapping[0])) individuals.add(vimd.getIndividual());
                        for(Object filler : (Set<Object>)(mapping[1])) individuals.add(img.object2OWL(filler).getIndividual());
                        setObjectProperties(individual, property, individuals);}
                    else {
                        IndividualWrapper vimd = img.getIWR(value);
                        OWLNamedIndividual filler  = (vimd == null) ? 
                                img.object2OWL(value).getIndividual() : vimd.getIndividual();
                        setObjectProperty(individual, property, filler);
                 }}}}}
    

    /** returns the sorted names of the OWLEntites. */
    static String[] getEntityNames(Collection<? extends OWLEntity> entities) {
        int length = entities.size();
        String[] names = new String[length];
        int i = 0;
        for(OWLEntity entity : entities) {
            names[i] = entity.getIRI().getFragment();
            ++i;}
        Arrays.sort(names);
        return names;}
 
    
    /** returns a string representation of the ontology manager. */
    @Override
    public String toString() {
        String blanks = "   ";
        String s = "Ontology Manager " + getName() + "\n";
        if(backgroundOntology != null) 
            s += blanks + "Background Ontology: " + getBackgroundOntologyIRI().toString() + "\n";
        if(extensionOntology != null) 
            s += blanks + "Extension Ontology:  " + getExtensionOntologyIRI().toString() + "\n";
        if(reasoner != null)
            s += blanks + "Reasoner:            " + reasoner.getReasonerName()+ "\n";
        s += blanks + "Ontology Manager:    " + ontologyManager.toString()+ "\n";
        s += blanks + "Data Factory:        " + dataFactory.toString()+ "\n";
        if(!changeAxioms.isEmpty()) 
            s += blanks + "Change Axioms:\n" + changeAxiomsString();
        if(!nameCounter.isEmpty()) {
            s += blanks + "Name counters: ";
            for(Entry<OWLClass,Integer> entry : nameCounter.entrySet())
                s += blanks + blanks + bidiShortFormProvider.getShortForm(entry.getKey()) + ": " +
                        entry.getValue().toString() + ", ";
            s += "\n";}
        s += blanks + "Datatype mapping:\n" + datatypeMappingString(blanks+blanks);  
        return s;
    }
         
    /** disposes of the reasoner */
    @Override
    public void finalize() {
        if(reasoner != null) reasoner.dispose();
    }
    
    /** This class helps keeping track of EquivalentClass annotations of Java classes.
     * The EquivalentClass annotation is a class expression string which can be parsed only
     * after all other axioms have been inserted into the ontology.
     */
    private class Equivalence {
        private Class jcl;
        private OWLClass ocl;
        private String classExpressionString;
        private Set<OWLClassExpression> superclasses;
        
        /** This represents an OWL equivalence
         *  ocl == classExpressionString and superclass_1 and ... and superclass_n.
         */
        Equivalence(Class jcl, OWLClass ocl, String classExpressionString, Set<OWLClassExpression> superclasses) {
            this.jcl = jcl;
            this.ocl = ocl;
            this.classExpressionString = classExpressionString;
            this.superclasses = superclasses;}
        
        /** inserts the equivalent class axiom into the changeAxioms */
       void addEquivalentClassesAxiom() {
            OWLClassExpression clex = null;
            try {clex = parseClassExpression(classExpressionString);} 
            catch (ParserException ex) {
                logger.error("Syntax error in class expression ",Logging.Origin.USAGE,clex, jcl.getName(), ex);
                return;}
            superclasses.add(clex);
            clex = dataFactory.getOWLObjectIntersectionOf(superclasses);
            changeAxioms.add(new AddAxiom(extensionOntology,
                    dataFactory.getOWLEquivalentClassesAxiom(ocl,clex)));}
    }
    
    
    
}
