/*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */

package Java2OWL;

import java.lang.reflect.Method;
import java.util.*;
import org.semanticweb.owlapi.model.*;

/** This is an auxiliary class for the PropertyWrapper. 
 *  <p>Since different getter methods in different classes can be mapped to the same OWL property,
 *  this data structure stores the information extracted from the particular getter method of a particular class.
 *
 * @author ohlbach
 */
class PropertyWrapperComponent {
    Class ownerClass;
    Method getter;
    Method setter = null;
    Method adder = null;
    Method remover = null;
    Method clearer = null;
    Method getKey = null;
    Class rangeClass = null;
    Class keyClass = null;
    boolean isFunctional = false;
    boolean isMapper = false;
    boolean isDataProperty = false;
    boolean isTotal = false;
    int atleast = -1;
    int atmost = -1;
    Map<String,Class[][]> types = new HashMap<String,Class[][]>();
    Map<String,Boolean> exactLength = new HashMap<String,Boolean>();
  
    
    /** constructs a property wrapper component for a getter method in a given class. */
    PropertyWrapperComponent(Class ownerClass, Method getter) {
        this.ownerClass = ownerClass;
        this.getter = getter;}
    
    Map<OWLProperty,Boolean> axiomsAdded = new HashMap<OWLProperty,Boolean>();
    
    /** adds all the axioms which depend on the domain to the axioms. */
    void addOWLAxioms(ArrayList<OWLAxiom> axioms, OWLProperty property, J2OClassManager classManager) {
        if(isMapper) {if(axiomsAdded.get(property) != null) return;
                      axiomsAdded.put(property,true);}
        OWLDataFactory dataFactory = J2OOntologyManager.dataFactory;
        OWLClass domain = classManager.getOWLClass(ownerClass,false,false);
        if(isDataProperty) { 
            OWLDataProperty pr = (OWLDataProperty)(property);
            OWLDataRange drange = J2OOntologyManager.toOWLDatatype.get(rangeClass);
            if(isFunctional) {
                if(isTotal) {axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                        dataFactory.getOWLDataSomeValuesFrom(pr, drange)));}}
                else {
                    if(atleast == atmost && atleast > 0) {
                        axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                            dataFactory.getOWLDataExactCardinality(atleast, pr)));}
                    else {
                        if(atleast > 0) {
                            axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                dataFactory.getOWLDataMinCardinality(atleast, pr)));}
                        if(atmost > 0) {
                            axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                dataFactory.getOWLDataMaxCardinality(atmost, pr)));}}}}
        
        else { // the property is an object-property.
            OWLObjectProperty pr = (OWLObjectProperty)property;
            // now we treat the atleast/atmost constraints.
                OWLClass range  = classManager.getOWLClass(rangeClass,false,false);
                if(isFunctional) {
                    if(isTotal) {axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                            dataFactory.getOWLObjectSomeValuesFrom(pr, range)));}}
                else {
                    if(atleast == atmost && atleast > 0) {
                        axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                dataFactory.getOWLObjectExactCardinality(atleast, pr)));}
                    else {
                        if(atleast > 0) {
                            axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                    dataFactory.getOWLObjectMinCardinality(atleast, pr,range)));}
                        if(atmost > 0) {
                            axioms.add(dataFactory.getOWLSubClassOfAxiom(domain, 
                                    dataFactory.getOWLObjectMaxCardinality(atmost, pr,range)));}}}}
        }

    
    
    /** returns a summary of the data. */
    @Override
    public String toString() {
        String s = "";
        if(setter != null)      s += "Setter: " + setter.getName()+"\n";
        if(adder != null)       s += "Adder: " + adder.getName()+"\n";
        if(remover != null)     s += "Remover: " + remover.getName()+"\n";
        if(clearer != null)     s += "Clearer: " + clearer.getName()+"\n";
        if(getKey != null)      s += "getKey: " + getKey.getName()+"\n";
        if(keyClass != null)    s += "Key Class: " + keyClass.getName()+"\n";
        if(rangeClass != null)  s += "Range Class: " + rangeClass.getName()+"\n";
        String p = "\nProperties:  " + (isFunctional ? "functional, " : "relational, ");
        if(isMapper) p += "mapper, ";
        p += isDataProperty ? "data property, " : "object property, ";
        if(isTotal) p += "total, ";
        if(atleast > 0) p += "atleast " + atleast + ", ";
        if(atmost > 0) p += "atmost " + atmost;
        
        String t = "\nTypes:\n";
        for(String accessor : types.keySet()) {
            t += accessor + ": exact length = " + exactLength.get(accessor) + "\n";
            for(Class[] jcls : types.get(accessor)) { 
                t += "        ";
                for(Class jcl : jcls) t += jcl.getSimpleName() + ", ";
                t += "\n";}}
        
        
        
        return "PropertyWrapperComponent for class: " + ownerClass.getName() + " and getter " + getter.getName() +":\n"
                + Utilities.adjustPairs(s, ":") + p + t;
    } 
            
}
